#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

import os
import sys

from srblib import Colour

from . import __version__, __mod_name__
from .scheduler import Scheduler
from .parser import get_parser
from .configurations import default_output_xlsx_path
from .verifier import Verifier

def main():
    args = get_parser()
    if args.version:
        print(__mod_name__+'=='+__version__)
        sys.exit()

    if args.vr:
        Verifier.verify_room_list(args.vr)
        sys.exit()
    if args.vs:
        Verifier.verify_schedule_list(args.vs)
        sys.exit()
    if args.vt:
        Verifier.verify_teachers_list(args.vt)
        sys.exit()
    if args.vw:
        Verifier.verify_work_ratio(args.vw)
        sys.exit()

    if args.seed <= 0:
        Colour.print('seed value should be a positive integer, got : ' + str(args.seed),Colour.RED)
        sys.exit(1)

    global default_output_xlsx_path
    if args.output: default_output_xlsx_path = args.output

    if args.reserved < 0:
        Colour.print('Reserved number should be a non-negative integer, got : ' + str(args.reserved), Colour.RED)
        sys.exit(1)

    try:
        scheduler = Scheduler(int(args.seed),int(args.reserved))
        if args.debug: scheduler.debug = True
        if args.info: scheduler.print_info = True
        scheduler._configure_paths() # done manually
        res = scheduler.compileall()
        if not res:
            Colour.print('Error during compilation',Colour.RED)
            print(res)
            sys.exit(1)
        scheduler.schedule(default_output_xlsx_path)
        Colour.print('Output written to : ' + Colour.END + default_output_xlsx_path, Colour.BLUE)
    except KeyboardInterrupt:
        Colour.print('Exiting on KeyboardInterrupt ...',Colour.YELLOW)

if(__name__=="__main__"):
    main()



import os
import sys
import json

from srblib import abs_path
from srblib import Colour
from srblib import SrbJson
from srblib import Tabular
from srblib import debug

from .priority_queue import PriorityQueue
from .session import Session
from .teacher import Teacher
from .util import randomize, fabricate
from .verifier import Compiler
from .work_ratio import WorkRatio

class Scheduler:
    def __init__(self,seed=5,reserved=0,room_list=None,teachers_list=None,schedule_list=None,work_ratio=None):
        self.debug = False
        self.seed = int(seed)
        self.reserved = int(reserved)
        self.room_list = Scheduler._verified_file(room_list)
        self.teachers_list = Scheduler._verified_file(teachers_list)
        self.schedule_list = Scheduler._verified_file(schedule_list)
        self.work_ratio = Scheduler._verified_file(work_ratio)
        self.workratio = None
        # self._configure_paths() # called manually
        self.print_info = False # if true then we will print tmap

    def _configure_paths(self):
        '''
        to be called manually, not to be used in library
        '''
        from .configurations import config_json
        def configure_path(var_name):
            if not getattr(self,var_name):
                if not config_json[var_name]:
                    path = input('Please provide ' + var_name + ' file : ')
                    path = Scheduler._verified_file(path)
                    config_json[var_name] = path
                    setattr(self,var_name,config_json[var_name])
                else:
                    setattr(self,var_name,config_json[var_name])
            else:
                config_json[var_name] = getattr(self,var_name)
        var_list = ['room_list','teachers_list','schedule_list','work_ratio']
        for var_name in var_list:
            configure_path(var_name)
            Colour.print('Using ' + var_name + ' : ' + Colour.END + getattr(self,var_name), Colour.GREEN)
        Colour.print('Using seed value : ' + Colour.END + str(self.seed), Colour.GREEN)
        Colour.print('Using reserved value : ' + Colour.END + str(self.reserved), Colour.GREEN)


    def compileall(self):
        res = Compiler.compile_room_list(self.room_list)
        if not res: return res
        res = Compiler.compile_work_ratio(self.work_ratio)
        if not res: return res
        res = Compiler.compile_teachers_list(self.teachers_list)
        if not res: return res
        res = Compiler.compile_schedule_list(self.schedule_list)
        if not res: return res
        res = Compiler.crosscompile(self.teachers_list,
                                    self.room_list,
                                    self.schedule_list,
                                    self.work_ratio
                                    )
        return res



    def schedule(self,output_path):
        self.workratio = WorkRatio(self.work_ratio)
        teachers_list = Teacher.get_teachers(self.teachers_list,self.workratio)
        session_list = Session.get_sessions(self.schedule_list,self.room_list,self.workratio)

        # for scheduling reserved using -r option
        self.schedule_reserved(teachers_list,session_list)

        # fake run
        self.generate_duties(teachers_list,session_list)

        # round_n_robin algorithm
        self.round_n_robin(teachers_list,session_list)

        # finished processing ... finally create output
        self.dump_output(teachers_list,session_list,output_path)


    def round_n_robin(self,teachers_list,session_list):
        '''
        Srb's round-n-robin algorithm
        '''
        sorted_teachers_list = teachers_list[:]
        sorted_teachers_list.sort(key=lambda x: int(x.rank))
        session_pq = PriorityQueue(fabricate(session_list,self.seed))

        def debug_pq(pq):
            if not self.debug: return
            need = 0
            for item in pq._data:
               session = item[1]
               for room in session.room_list:
                   need += room.unfilled()
            print('Session PQ has rooms: ',need)

        debug_how = 0
        for session in session_list:
            for room in session.room_list:
                debug_how += room.teachers

        for teacher in sorted_teachers_list:
            done_list = []
            for _ in range(teacher.duties):
                session = session_pq.pop()
                while(session.name in teacher.alloted_res):
                    if(self.debug): print('skipped session\n',session)
                    done_list.append(session)
                    session = session_pq.pop()

                if(self.debug): print(session)
                done_list.append(session)

                room = session.room_pq.pop()
                room.teachers_alloted.append(teacher)
                teacher.alloted[session.name] = room.name
                session.remaining -= 1
                debug_how -= 1
                if room.unfilled() > 0: session.room_pq.push(room)
            for session in done_list:
                if session.remaining > 0: session_pq.push(session)
            if(self.debug):print(teacher)
            if(self.debug):print('debug_how',debug_how)
            if(self.debug):debug_pq(session_pq)
            if(self.debug):print()




    def dump_output(self,teachers_list,session_list,output_path):
        matrix = [["Name of Faculty Member","Info"]]
        if debug: matrix[0].extend(['rank','total','res','credits']) # srbdebug

        dmap = {} # map to contain count of avg duties of some rank
        rmap = {} # map to contain count of teachers of some rank
        for teacher in teachers_list:
            rmap[teacher.rank] = 0
            dmap[teacher.rank] = 0
        for teacher in teachers_list:
            rmap[teacher.rank] += 1
            dmap[teacher.rank] += teacher.duties
        def divide(a,b):
            a = int((a*1000)/b)
            return a/1000
        for rank in rmap:
            dmap[rank] = divide(dmap[rank],rmap[rank])

        tmap = {} # map to contain count of room types
        for session in session_list:
            matrix[0].append(session.name)
            for room in session.room_list:
                if room.get_type() in tmap:
                    tmap[room.get_type()] += 1
                else:
                    tmap[room.get_type()] = 1

        matrix[0].append("Total")

        for teacher in teachers_list:
            teachers_row = [teacher.name,teacher.info]
            if teacher.duties != len(teacher.alloted) - len(teacher.alloted_res):
                print('ERROR: teacher unable to get enough slots as anticipated')
                print(teacher)
                sys.exit(1)
            if debug:
                teachers_row.extend([
                    teacher.rank,
                    len(teacher.alloted),
                    len(teacher.alloted_res),
                    int(teacher._credits)
                ])
            for session in session_list:
                if session.name in teacher.alloted: teachers_row.append(teacher.alloted[session.name])
                else: teachers_row.append('-')
            teachers_row.append(len(teacher.alloted))
            matrix.append(teachers_row)

        lmap = json.dumps(rmap,indent=3,sort_keys=True)
        Colour.print('rank count : ',Colour.CYAN,end='')
        Colour.print(lmap,Colour.GREEN)

        lmap = json.dumps(dmap,indent=3,sort_keys=True)
        Colour.print('average duties : ',Colour.CYAN,end='')
        Colour.print(lmap,Colour.GREEN)

        lmap = json.dumps(tmap,indent=3,sort_keys=True)
        Colour.print('type of rooms : ',Colour.CYAN,end='')
        Colour.print(lmap,Colour.GREEN)

        if self.print_info:
            matrix.extend([[],[],['rank','count','avg duties']])
            for key in sorted(rmap):
                matrix.append([key,rmap[key],dmap[key]])

            matrix.extend([[],[],['type of room','number']])
            for key in sorted(tmap):
                matrix.append([key,tmap[key]])

        sheet = Tabular(matrix)
        sheet.write_xls(output_path)


    def generate_duties(self,teachers_list,session_list):
        '''
        Fake run: just to determine teacher duties number
        '''
        teachers_pq = PriorityQueue(randomize(teachers_list,self.seed))
        for session in session_list:
            done_list = []
            for room in session.room_list:
                for _ in range(room.teachers):
                    teacher = teachers_pq.pop()
                    while session.name in teacher.alloted_res: # teacher should not get sametime res and room
                        done_list.append(teacher)
                        teacher = teachers_pq.pop()
                    teacher._credits += self.workratio.credits_calc(teacher.rank)
                    teacher.duties += 1
                    done_list.append(teacher)
            for teacher in done_list:
                teachers_pq.push(teacher)


    def schedule_reserved(self,teachers_list,session_list):
        '''
        this res is triggered using -r, it will be alloted according to rank
        smaller ranks will get more chances to be reserved
        there is another way, that is using creating a RES room in roomlist (not recommended)
        both differ by name Res for this one RES for that one

            reason for not being recommended theh roomlist way of reserving:
                1. equal credits for normal and reserved room
                2. unequal distribution of reserved seats
                3. no priority to better rank teachers.
                4. makes priority queue difficult to schedule if reserved room contain more teachers than other rooms
        '''
        teachers_list_res = Teacher.get_teachers(self.teachers_list,self.workratio) # create totally new one
        for i in range(len(teachers_list_res)):
            teachers_list_res[i]._credits = 99999 if teachers_list_res[i].rank == 0 else teachers_list_res[i].rank
        teachers_reserved_pq = PriorityQueue(randomize(teachers_list_res,self.seed+1))
        max_rank = Scheduler._get_max_rank(teachers_list)
        for session in session_list: # reserve
            done_list = []
            for j in range(self.reserved):
                teacher = teachers_reserved_pq.pop()
                teacher._credits += self.workratio.credits_calc(max_rank-teacher.rank + 1)
                done_list.append(teacher)
                teachers_list[teacher.idd-2].alloted[session.name] = 'Res'
                teachers_list[teacher.idd-2].alloted_res.add(session.name)
                teachers_list[teacher.idd-2]._credits += self.workratio.credits_calc(teacher.rank) / 2 # should get half credits for reserved seat
            for teacher in done_list:
                teachers_reserved_pq.push(teacher)


    @staticmethod
    def _verified_file(file_path):
        if not file_path:
            return None
        file_path = abs_path(file_path)
        if not os.path.isfile(file_path):
            Colour.print('No such file : ' + Colour.END + file_path,Colour.RED)
            sys.exit()
        return file_path

    @staticmethod
    def _get_max_rank(teachers_list):
        max_rank = 0
        for teacher in teachers_list:
            max_rank = max(max_rank,teacher.rank)
        return max_rank





import os
import sys

from srblib import show_dependency_error_and_exit

try:
    import argparse
    from argcomplete import autocomplete
except:
    show_dependency_error_and_exit()

def get_parser():
    def _is_valid_file(parser, arg):
        if not os.path.isfile(arg):
            parser.error("The file %s does not exist!" % arg)
        else:
            return arg

    parser = argparse.ArgumentParser()
    parser.add_argument("-v","--version",
                        action='store_true',
                        help='Display version number')
    parser.add_argument("-o","--output",
                        # type=lambda x: _is_valid_file(parser,x),
                        type=str,
                        help='Output file name, default output.xlsx')
    parser.add_argument("-s","--seed",
                        default=5,
                        type=int,
                        help='seed value for randomness')
    parser.add_argument("-r","--reserved",
                        default=0,
                        type=int,
                        help='reserved number of seats for each session')
    parser.add_argument("-d","--debug",
                        action='store_true',
                        help='print debug info')
    parser.add_argument("-i","--info",
                        action='store_true',
                        help='print info in excel')

    verify_parser = parser.add_mutually_exclusive_group()
    verify_parser.add_argument("-vr",
                        type=lambda x: _is_valid_file(parser,x),
                        help='verify room_list file')
    verify_parser.add_argument("-vs",
                        type=lambda x: _is_valid_file(parser,x),
                        help='verify schedule_list file')
    verify_parser.add_argument("-vt",
                        type=lambda x: _is_valid_file(parser,x),
                        help='verify teachers_list file')
    verify_parser.add_argument("-vw",
                        type=lambda x: _is_valid_file(parser,x),
                        help='verify work_ratio file')

    autocomplete(parser)
    return parser.parse_args()



from srblib import Tabular

import copy

from .verifier import Verifier

class Room:
    def __init__(self,name,teachers,capacity=0,**kwargs):
        '''
        format:
            required positional params:
                name - name of room
                teachers - number of teachers
            optional parameters:
                capacity - capacity of students (default)
            extra:
                kwargs - may include any property
        '''
        self.name = name
        self.teachers = int(teachers)
        self.capacity = int(capacity)
        self.extra = list(kwargs.keys())
        self.__dict__.update(kwargs)
        self.teachers_alloted = []
        self._credits = -1 # weitage of room
        self.workratio = None # WorkRatio object

    def unfilled(self):
        return self.teachers - len(self.teachers_alloted)

    def filled(self):
        return len(self.teachers_alloted)

    def empty(self):
        if len(self.teachers_alloted) == 0:
            return True
        return False

    def get_credits(self):
        self._credits = 0
        for teacher in self.teachers_alloted:
            self._credits += self.workratio.credits_calc(teacher.rank)
        return self._credits

    def __eq__(self,obj):
        if not self < obj and not obj < self:
            return True
        return False

    def __lt__(self,obj):
        if self.empty() and not obj.empty(): return True # special case when self is totally empty it must get highest priority
        if self.unfilled() == obj.unfilled():
            if self.filled() == obj.filled():
                return self.get_credits() < obj.get_credits()
            else:
                return self.filled() < obj.filled()
        return self.unfilled() > obj.unfilled()

    def copy(self):
        ret = copy.deepcopy(self)
        return ret

    def get_type(self):
        arr = []
        for teacher in self.teachers_alloted:
            arr.append(teacher.rank)
        arr.sort()
        return arr.__str__()

    def __str__(self):
        a = [
                ["name",self.name],
                ["teachers-required",self.teachers],
                ["capacity",self.capacity],
                ["teachers-alloted"," ".join([teacher.name for teacher in self.teachers_alloted])]
            ]
        for key in self.extra:
            a.append([key,getattr(self,key)])

        a = Tabular(a)
        return a.__str__()

    @staticmethod
    def get_rooms(matrix,workratio):
        '''
        input should be a Tabular object, or a path
        workratio is object of WorkRatio
        '''
        if type(matrix) is str:
            Verifier.verify_room_list(matrix)
            temp = Tabular()
            temp.load_xls(matrix,strict=True)
            matrix = temp
        ret = []
        header = matrix[0]
        cols = len(header)
        if(cols < 2):
            raise Exception('too few columns')

        matrix = matrix[1:]
        for row in matrix:
            capacity = 0
            kwargs = dict()
            if(cols >= 3): capacity = row[2]
            if(cols > 3):
                i = 3
                while i < cols:
                    kwargs[header[i]] = row[i]
                    i += 1
            temp = Room(row[0],row[1],capacity,**kwargs)
            temp.workratio = workratio
            ret.append(temp)

        return ret



import sys
from srblib import Tabular

from .util import Response

class Compiler:
    def safefun(fun):
        def inner(matrix):
            if type(matrix) == str:
                matrix = Compiler._read(matrix)
            if type(matrix) is Response: return matrix
            return fun(matrix)
        return inner

    @staticmethod
    @safefun
    def compile_room_list(matrix):
        r = Response()
        header = matrix[0]
        cols = len(header)

        if(cols < 2):
            r.addMessage("In room_list : " + str(1) + " :")
            r.addMessage("Too few columns, require at least 2 cols for name and teachers")
            r.addMessage("Found " + str(cols) + " cols")
            r.addMessage("")
            return r

        found = set()
        rownum = 1
        for row in matrix[1:]:
            rownum += 1
            try:
                int_val = int(row[1])
                if int_val < 0:
                    r.addMessage("In room_list : " + str(rownum) + " : " + str(2) + " : ")
                    r.addMessage("Second column should contain non-negative integer value")
                    r.addMessage("Found : '" + str(row[1]) + "'")
                    r.addMessage("")
            except:
                r.addMessage("In room_list : " + str(rownum) + " : " + str(2) + " :")
                r.addMessage("Second column should contain non-negative integer value")
                r.addMessage("Found : '" + str(row[1]) + "'")
                r.addMessage('\n')


            if row[0] in found:
                r.addMessage("In room_list : " + str(rownum) + " : " + str(1) + " :")
                r.addMessage("First column should contain unique value")
                r.addMessage("Found : '" + str(row[0]) + "' twice")
                r.addMessage("")

            found.add(row[0])

        return r

    @staticmethod
    @safefun
    def compile_work_ratio(matrix):
        r = Response()
        header = matrix[0]
        cols = len(header)

        if(cols < 2):
            r.addMessage("In work_ratio : " + str(1) + " :")
            r.addMessage("Too few columns, require at least 2 cols for rank and work_ratio")
            r.addMessage("Found " + str(cols) + " cols")
            r.addMessage("")
            return r

        found = set()
        rownum = 1
        for row in matrix[1:]:
            rownum += 1
            try:
                int_val = int(row[0])
                if int_val < 0:
                    r.addMessage("In work_ratio : " + str(rownum) + " : " + str(1) + " :")
                    r.addMessage("First(rank) column should contain non-negative integer value")
                    r.addMessage("Found : '" + str(row[0]) + "'")
                    r.addMessage("")
            except:
                r.addMessage("In work_ratio : " + str(rownum) + " : " + str(1) + " :")
                r.addMessage("First(rank) column should contain non-negative integer value")
                r.addMessage("Found : '" + str(row[0]) + "'")
                r.addMessage("")


            try:
                int_val = float(row[1])
                if int_val < 0:
                    r.addMessage("In work_ratio : " + str(rownum) + " : " + str(2) + " :")
                    r.addMessage("Second(ratio) column should contain non-negative float value")
                    r.addMessage("Found : '" + str(row[1]) + "'")
                    r.addMessage("")
            except:
                r.addMessage("In work_ratio : " + str(rownum) + " : " + str(2) + " :")
                r.addMessage("Second(ratio) column should contain non-negative float value")
                r.addMessage("Found : '" + str(row[1]) + "'")
                r.addMessage("")


            if row[0] in found:
                r.addMessage("In work_ratio : " + str(rownum) + " : " + str(1) + " :")
                r.addMessage("First column should contain unique value")
                r.addMessage("Found : '" + str(row[0]) + "' twice")
                r.addMessage("")

            found.add(row[0])

        return r


    @staticmethod
    @safefun
    def compile_teachers_list(matrix):
        r = Response()
        header = matrix[0]
        cols = len(header)

        if(cols < 2):
            r.addMessage("In teachers_list : " + str(1) + " :")
            r.addMessage("Too few columns, require at least 2 columns for name and rank")
            r.addMessage("Found " + str(cols) + " cols")
            r.addMessage("")
            return r

        found = set()
        rownum = 1
        for row in matrix[1:]:
            rownum += 1
            try:
                int_val = int(row[1])
                if int_val < 0:
                    r.addMessage("In teachers_list : " + str(rownum) + " : " + str(2) + " :")
                    r.addMessage("Second(rank) column should contain non-negative integer value")
                    r.addMessage("Found : '" + str(row[1]) + "'")
                    r.addMessage("")
            except:
                r.addMessage("In teachers_list : " + str(rownum) + " : " + str(2) + " :")
                r.addMessage("Second(rank) column should contain non-negative integer value")
                r.addMessage("Found : '" + str(row[1]) + "'")
                r.addMessage("")


            if row[0] in found:
                r.addMessage("In teachers_list : " + str(rownum) + " : " + str(1) + " :")
                r.addMessage("First(name) column should contain unique value")
                r.addMessage("Found : '" + str(row[0]) + "' twice")
                r.addMessage("")

            found.add(row[0])
        return r


    @staticmethod
    @safefun
    def compile_schedule_list(matrix):
        r = Response()
        header = matrix[0]
        cols = len(header)

        if(cols < 2):
            r.addMessage("In schedule_list : " + str(1) + " : ")
            r.addMessage("Too few columns, require at least 2 columns for session_name and rooms")
            r.addMessage("Found " + str(cols) + " cols")
            r.addMessage("")
            return r

        found = set()
        colnum = 0
        for head in header:
            colnum+=1
            if head in found:
                r.addMessage("In schedule_list : " + str(1) + " : " + str(colnum) + " :")
                r.addMessage("Room names should be value")
                r.addMessage("Found : '" + str(head) + "' twice")
                r.addMessage("")
            found.add(head)

        found = set()
        rownum = 1
        for row in matrix[1:]:
            rownum += 1
            if row[0] in found:
                r.addMessage("In schedule_list : " + str(rownum) + " : " + str(1) + " :")
                r.addMessage("First(session_name) column should have unique value")
                r.addMessage("Found : '" + str(row[0]) + "' twice")
                r.addMessage("")

            colnum = 1
            for cell in row[1:]:
                colnum+=1
                if cell and cell not in ('Y','y','N','n'):
                    r.addMessage("In schedule_list : " + str(rownum) + " : " + str(colnum) + " :")
                    r.addMessage("Cell(other than session-name) can be blank, Y, y, N or n only")
                    r.addMessage("Found : '" + str(row[0]) + "' twice")
                    r.addMessage("")

            found.add(row[0])
        return r


    @staticmethod
    def crosscompile(teachers_list,room_list,schedule_list,work_ratio):
        '''
        only to be called after compilation is done(all files are OK)
        '''
        teachers_list = Compiler._read(teachers_list)
        work_ratio = Compiler._read(work_ratio)
        room_list = Compiler._read(room_list)
        schedule_list = Compiler._read(schedule_list)

        r = Response()

        ratio_rank = set()
        for row in work_ratio[1:]: ratio_rank.add(int(row[0]))
        rank_list = set()
        for row in teachers_list[1:]: rank_list.add(int(row[1]))
        for rank in rank_list:
            if rank == 0: continue
            if rank not in ratio_rank:
                r.addMessage("'" + str(rank) + "' not present in work_ratio file, (present in teachers_list)")
                r.addMessage("")

        rooms = set()
        for room in room_list[1:]: rooms.add(str(room[0]))
        for room in schedule_list[0][1:]:
            if room not in rooms:
                r.addMessage("'" + str(room) + "' not present in room_list file, (present in schedule_list)")
                r.addMessage("")

        return r

    @staticmethod
    def _read(file_path):
        try:
            matrix = Tabular()
            matrix.load_xls(file_path,strict=True)
            return matrix.matrix
        except Exception as e:
            r = Response(e)
            r.addMessage("[Exception] Unable to load file")
            r.addMessage("")
            return r



class Verifier:
    '''
    this class is for commandline purpose, should not be used as part of module
    '''
    def dieifbad(fun):
        def inner(*args,**kwargs):
            ret = fun(*args,**kwargs)
            if not ret.safe():
                print(ret)
                sys.exit(1)
        return inner

    @staticmethod
    @dieifbad
    def verify_room_list(file_path):
        return Compiler.compile_room_list(file_path)

    @staticmethod
    @dieifbad
    def verify_teachers_list(file_path):
        return Compiler.compile_teachers_list(file_path)

    @staticmethod
    @dieifbad
    def verify_schedule_list(file_path):
        return Compiler.compile_schedule_list(file_path)

    @staticmethod
    @dieifbad
    def verify_work_ratio(file_path):
        return Compiler.compile_work_ratio(file_path)

    @staticmethod
    @dieifbad
    def crossverify(teachers_list,room_list,schedule_list,work_ratio):
        return Compiler.crosscompile(teachers_list,room_list,schedule_list,work_ratio)



import heapq

class PriorityQueue:
    def __init__(self, data=None, key=lambda x:0):
        self.empty = True
        self.key = key
        if data:
            self._data = [(key(item), item) for item in data]
            heapq.heapify(self._data)
            self.empty = False
        else:
            self._data = []

    def push(self, item):
        heapq.heappush(self._data, (self.key(item), item))
        self.empty = False

    def pop(self):
        if(len(self._data) == 0): raise Exception('queue empty !! unable to pop')
        ret = heapq.heappop(self._data)[1]
        if(len(self._data) == 0): self.empty = True
        return ret

    def top(self):
        if(len(self._data) == 0): raise Exception('queue empty !! unable to top')
        return self._data[0][1]

    def __len__(self):
        return len(self._data)

    def __str__(self):
        arr = [x[1] for x in self._data]
        return arr.__str__()



from srblib import Tabular
from srblib import debug

class Teacher:
    def __init__(self,idd,name,rank,info="",**kwargs):

        '''
        format:
            required positional params:
                idd - unique id (ususlly not by user)
                name - name of teacher
                rank - rank of teacher
            optional parameters:
                info - any extra info
            extra:
                kwargs - may include any property
        '''
        self.idd = idd
        self.name = name
        self.rank = int(rank)
        self.info = info
        self.subclass = "" # extract out if something after name in brackets
        self.info = info
        self.extra = list(kwargs.keys())
        self.__dict__.update(kwargs)
        self._credits = 0
        self.alloted = {}
        self.alloted_res = set()
        self.duties = 0 # count of real duties, except reserved ones

    def __str__(self):
        a = [
                ["id",self.idd],
                ["name",self.name],
                ["rank",self.rank],
                ["info",self.info],
                ["duties",self.duties],
                ["alloted_res","\n".join([str(x) for x in self.alloted_res])],
                ["alloted","\n".join([str((x,self.alloted[x])) for x in self.alloted])],
            ]
        if debug: a.append(["credits",self._credits])
        for key in self.extra:
            a.append([key,getattr(self,key)])

        a = Tabular(a)
        return a.__str__()

    def __lt__(self,obj):
        if self._credits == obj._credits:
            return self.rank > obj.rank
        return self._credits < obj._credits

    @staticmethod
    def get_teachers(matrix,workratio):
        '''
        input should be a Tabular object, or a path
        workratio is WorkRatio object
        '''
        if type(matrix) is str:
            temp = Tabular()
            temp.load_xls(matrix,strict=True)
            matrix = temp
        ret = []
        header = matrix[0]
        cols = len(header)
        if(cols < 2):
                raise Exception('too few columns')
        idd = 2
        for row in matrix:
            info = ""
            kwargs = dict()
            if(cols >= 3): info = row[2]
            if(cols > 3):
                i = 3
                while i < cols:
                    kwargs[header[i]] = row[i]
                    i += 1
            temp = Teacher(idd,row[0],row[1],info,**kwargs)
            temp._credits = workratio.credits_calc(temp.rank) / 10 # just for initial sort
            ret.append(temp)
            idd += 1

        return ret




import sys

from srblib import Colour
from srblib import Tabular

from .room import Room
from .priority_queue import PriorityQueue
from .verifier import Verifier

class Session:
    def __init__(self,name,room_list):
        self.name = name
        self.room_list = room_list
        self.remaining = 0 # slots not filled
        for room in room_list:
            self.remaining += room.teachers
        self.room_pq = PriorityQueue(self.room_list)

    def __str__(self):
        a = [[self.name]]
        a[0].append(self.remaining)
        a[0].append(self.room_pq.top().name)
        a[0].append(self.room_pq.top().unfilled())
        a[0].append(self.room_pq.top().filled())
        a[0].append(self.room_pq.top().get_type())
        for room in self.room_list:
            a[0].append(room.name + ' ' + str(room.teachers - len(room.teachers_alloted)))
        a = Tabular(a)
        return a.__str__()

    def __lt__(self,obj):
        self_top = self.room_pq.top()
        obj_top = obj.room_pq.top()
        if self_top.empty() and not obj_top.empty(): return True # special case, empty room should be filled first
        if self_top.unfilled() == obj_top.unfilled() and self_top.filled() == obj_top.filled():
            if self.remaining == obj.remaining:
                return self_top < obj_top
            else:
                return self.remaining > obj.remaining
        return self_top < obj_top

    @staticmethod
    def get_sessions(matrix,room_data,workratio):
        '''
        input matrix should be a Tabular object, or a path
        room_data should be a tabular object or a path
        workratio is obj of WorkRatio
        '''
        if type(matrix) is str:
            Verifier.verify_schedule_list(matrix)
            temp = Tabular()
            temp.load_xls(matrix,strict=True)
            matrix = temp
        if type(room_data) is str:
            room_data = Room.get_rooms(room_data,workratio)

        room_json = {}
        for room in room_data:
            room_json[room.name] = room

        ret = []
        header = matrix[0][1:] # remove first elem from header
        cols = len(header) + 1
        if(cols < 2):
                raise Exception('too few columns')
        for i in range(1,len(matrix)):
            row = matrix[i]
            room_list = []
            for room in header:
                need = row[room]
                if need == 'y' or need == 'Y':
                    if room in room_json:
                        room_list.append(room_json[room].copy())
                    else:
                        # raise Exception('Room ' +room+ ' not present in rooms file')
                        Colour.print('Room ' + room + ' not present in room list',Colour.RED)
                        sys.exit()

            temp = Session(row[0], room_list)
            ret.append(temp)

        return ret



import os

from srblib import abs_path
from srblib import SrbJson
from srblib import on_ci

config_file_path = abs_path('~/.config/exam_scheduler/config.json')
default_output_xlsx_path = abs_path('./output.xls')

_default_room_list = None
_default_teachers_list = None
_default_schedule_list = None
_default_work_ratio = None

if os.path.exists(abs_path('input/')):
    for fname in os.listdir(abs_path('input/')):
        fname = 'input/' + fname
        if not _default_room_list and 'room_list' in fname:
            _default_room_list = abs_path(fname)
        elif not _default_teachers_list and 'teachers_list' in fname:
            _default_teachers_list = abs_path(fname)
        elif not _default_schedule_list and 'schedule_list' in fname:
            _default_schedule_list = abs_path(fname)
        elif not _default_work_ratio and 'work_ratio' in fname:
            _default_work_ratio = abs_path(fname)

if on_ci and os.path.exists(abs_path('.ci/')):
    for fname in os.listdir(abs_path('.ci/')):
        fname = '.ci/' + fname
        if not _default_room_list and 'room_list' in fname:
            _default_room_list = abs_path(fname)
        elif not _default_teachers_list and 'teachers_list' in fname:
            _default_teachers_list = abs_path(fname)
        elif not _default_schedule_list and 'schedule_list' in fname:
            _default_schedule_list = abs_path(fname)
        elif not _default_work_ratio and 'work_ratio' in fname:
            _default_work_ratio = abs_path(fname)

for fname in os.listdir(abs_path('./')):
    if not _default_room_list and 'room_list' in fname:
        _default_room_list = abs_path(fname)
    elif not _default_teachers_list and 'teachers_list' in fname:
        _default_teachers_list = abs_path(fname)
    elif not _default_schedule_list and 'schedule_list' in fname:
        _default_schedule_list = abs_path(fname)
    elif not _default_work_ratio and 'work_ratio' in fname:
        _default_work_ratio = abs_path(fname)

_config_template = \
{
    'room_list':_default_room_list,
    'teachers_list':_default_teachers_list,
    'schedule_list':_default_schedule_list,
    'work_ratio':_default_work_ratio,
}

config_json = SrbJson(config_file_path,_config_template)

for var_name in ['room_list','teachers_list','schedule_list','work_ratio']:
    if config_json[var_name] and not os.path.isfile(config_json[var_name]):
        config_json[var_name] = None
    if not config_json[var_name]:
        config_json[var_name] = _config_template[var_name]



from random import Random
from srblib import Tabular

def fabricate(arr,seed=5):
    seed = int(seed)
    if seed >= len(arr)/2: seed = len(arr)//3
    ret = []
    i = seed
    while i > 0:
        j = i - 1
        while j < len(arr):
            ret.append(arr[j])
            j += seed
        i -= 1
    return ret

def randomize(arr,seed=None):
    r = Random()
    if seed: r.seed(str(seed))
    arr = fabricate(arr,7)
    n = len(arr)
    for i in range(0,n-1):
        j = r.randint(i,n-1)
        arr[i],arr[j] = arr[j],arr[i]
    for i in range(n-1,0,-1):
        j = r.randint(0,i)
        arr[i],arr[j] = arr[j],arr[i]
    return arr


class Response:
    def __init__(self,execption=None):
        self.__message = []
        self.exception = execption

    def addMessage(self, msg):
        self.__message.append(msg)
        return self

    def __str__(self):
        ans = ""
        # if len(self.__message) > 0:
            # ans += "[ERROR]\n\n"
        ans += '\n'.join(self.__message)
        # if len(self.__message) > 0:
            # ans += "\n[NOTE] Row nums are not exact, Empty rows are not counted\n"
        return ans

    def __bool__(self):
        return len(self.__message)==0 and not self.exception

    def safe(self):
        if self.exception:
            print(Tabular([[str(self)]]))
            raise self.exception
        return bool(self)

    def json(self):
        return {
            'message':str(self),
            'status':bool(self)
        }



from srblib import Tabular

class WorkRatio:
    def __init__(self,filepath):
        matrix = Tabular()
        matrix.load_xls(filepath,strict=True)
        matrix = matrix.matrix[1:]
        self.ratio = {}
        self.credits = {0:999999}
        for x in matrix:
            self.ratio[int(x[0])] = float(x[1])

        if not 1 in self.ratio: self.ratio[1] = 100
        for x in matrix:
            self.credits[int(x[0])] = self._credits_calc(int(x[0]))


    def _credits_calc(self,rank):
        return (100*self.ratio[1])/self.ratio[rank]

    def credits_calc(self,rank):
        rank = int(rank)
        return self.credits[rank]
